searchNodes=[{"ref":"Xema.html","title":"Xema","type":"module","doc":"A schema validator inspired by JSON Schema. All available keywords to construct a schema are described on page Usage. This module can be used to construct a schema module. Should a module contain multiple schemas the option multi: true is required. use Xema imports Xema.Builder and extends the module with the functions __MODULE__.valid?/2 __MODULE__.validate/2 __MODULE__.validate!/2 __MODULE__.cast/2 __MODULE__.cast!/2 The macro xema/2 supports the construction of a schema. After that the schema is available via the functions above. In a multi schema module a schema can be tagged with @default true and then called by __MODULE__.valid?/1 __MODULE__.validate/1 __MODULE__.validate!/1 __MODULE__.cast/1 __MODULE__.cast!/1 The functions with arity 1 are also available for single schema modules. Examples Sinlge schema module: iex&gt; defmodule SingleSchema do ...&gt; use Xema ...&gt; ...&gt; # The name :num is optional. ...&gt; xema :num, do: number(minimum: 1) ...&gt; end iex&gt; iex&gt; SingleSchema.valid?(:num, 6) true iex&gt; SingleSchema.valid?(5) true iex&gt; SingleSchema.validate(0) {:error, %Xema.ValidationError{ reason: %{minimum: 1, value: 0} }} iex&gt; SingleSchema.cast(&quot;5&quot;) {:ok, 5} iex&gt; SingleSchema.cast(&quot;-5&quot;) {:error, %Xema.ValidationError{ reason: %{minimum: 1, value: -5} }} Multi schema module: iex&gt; defmodule Schema do ...&gt; use Xema, multi: true ...&gt; ...&gt; @pos integer(minimum: 0) ...&gt; @neg integer(maximum: 0) ...&gt; ...&gt; @default true ...&gt; xema :user do ...&gt; map( ...&gt; properties: %{ ...&gt; name: string(min_length: 1), ...&gt; age: @pos ...&gt; } ...&gt; ) ...&gt; end ...&gt; ...&gt; xema :nums do ...&gt; map( ...&gt; properties: %{ ...&gt; pos: list(items: @pos), ...&gt; neg: list(items: @neg) ...&gt; } ...&gt; ) ...&gt; end ...&gt; end iex&gt; iex&gt; Schema.valid?(:user, %{name: &quot;John&quot;, age: 21}) true iex&gt; Schema.valid?(%{name: &quot;John&quot;, age: 21}) true iex&gt; Schema.valid?(%{name: &quot;&quot;, age: 21}) false iex&gt; Schema.validate(%{name: &quot;John&quot;, age: 21}) :ok iex&gt; Schema.validate(%{name: &quot;&quot;, age: 21}) {:error, %Xema.ValidationError{ reason: %{ properties: %{name: %{min_length: 1, value: &quot;&quot;}}} } } iex&gt; Schema.valid?(:nums, %{pos: [1, 2, 3]}) true iex&gt; Schema.valid?(:nums, %{neg: [1, 2, 3]}) false ```"},{"ref":"Xema.html#cast/3","title":"Xema.cast/3","type":"function","doc":"Converts the given data using the specified schema. Returns {:ok, result} or {:error, reason}. The result is converted and validated with the schema. Examples: iex&gt; schema = Xema.new({:integer, minimum: 1}) iex&gt; Xema.cast(schema, &quot;5&quot;) {:ok, 5} iex&gt; Xema.cast(schema, &quot;five&quot;) {:error, %Xema.CastError{ key: nil, path: [], to: :integer, value: &quot;five&quot; }} iex&gt; Xema.cast(schema, &quot;0&quot;) {:error, %Xema.ValidationError{ reason: %{minimum: 1, value: 0} }} Multiple types If for a value multiple types are defined the function used the result of the first successful conversion. Examples iex&gt; schema = Xema.new([:integer, :string, nil]) iex&gt; Xema.cast(schema, 5) {:ok, 5} iex&gt; Xema.cast(schema, 5.5) {:ok, &quot;5.5&quot;} iex&gt; Xema.cast(schema, &quot;5&quot;) {:ok, 5} iex&gt; Xema.cast(schema, &quot;five&quot;) {:ok, &quot;five&quot;} iex&gt; Xema.cast(schema, nil) {:ok, nil} iex&gt; Xema.cast(schema, [5]) {:error, %Xema.CastError{path: [], to: [:integer, :string, nil], value: [5]} } Cast with any_of, all_of, and one_of Schemas in a combiner will be cast independently one by one in reverse order. No CasteErrors will be raised for this schemata. Examples iex&gt; schema = Xema.new(any_of: [ ...&gt; [properties: %{a: :integer}], ...&gt; [properties: %{a: :string}] ...&gt; ]) iex&gt; Xema.cast(schema, %{a: 5}) {:ok, %{a: 5}} iex&gt; Xema.cast(schema, %{a: 5.5}) {:ok, %{a: &quot;5.5&quot;}} iex&gt; Xema.cast(schema, %{a: &quot;5&quot;}) {:ok, %{a: 5}} iex&gt; Xema.cast(schema, %{a: &quot;five&quot;}) {:ok, %{a: &quot;five&quot;}} iex&gt; Xema.cast(schema, %{a: [5]}) {:error, %Xema.ValidationError{ reason: %{ any_of: [ %{properties: %{a: %{type: :integer, value: [5]}}}, %{properties: %{a: %{type: :string, value: [5]}}} ], value: %{a: [5]} } } } Options With the option additional_properties: :delete additional properties will be deleted on cast. Additional properties will be deleted in schemas with additional_properties: false. Examples iex&gt; schema = Xema.new( ...&gt; properties: %{ ...&gt; a: [ ...&gt; properties: %{ ...&gt; foo: :integer ...&gt; }, ...&gt; additional_properties: false ...&gt; ], ...&gt; b: [ ...&gt; properties: %{ ...&gt; foo: :integer ...&gt; } ...&gt; ] ...&gt; } ...&gt; ) iex&gt; iex&gt; Xema.cast(schema, %{ ...&gt; a: %{foo: &quot;6&quot;, bar: &quot;7&quot;}, ...&gt; b: %{foo: &quot;6&quot;, bar: &quot;7&quot;}, ...&gt; }, additional_properties: :delete) {:ok, %{ a: %{foo: 6}, b: %{foo: 6, bar: &quot;7&quot;} }}"},{"ref":"Xema.html#cast!/3","title":"Xema.cast!/3","type":"function","doc":"Converts the given data using the specified schema. Returns the converted data or an exception."},{"ref":"Xema.html#new/2","title":"Xema.new/2","type":"function","doc":"This function creates the schema from the given data. Possible options: :loader - a loader for remote schemas. This option will overwrite the loader from the config. See Configure a loader to how to define a loader. inline - inlined all references in the schema. Default :true. Examples Simple schema: iex&gt; schema = Xema.new :string iex&gt; Xema.valid? schema, &quot;hello&quot; true iex&gt; Xema.valid? schema, 42 false Schema: iex&gt; schema = Xema.new {:string, min_length: 3, max_length: 12} iex&gt; Xema.valid? schema, &quot;hello&quot; true iex&gt; Xema.valid? schema, &quot;hi&quot; false Nested schemas: iex&gt; schema = Xema.new {:list, items: {:number, minimum: 2}} iex&gt; Xema.validate(schema, [2, 3, 4]) :ok iex&gt; Xema.valid?(schema, [2, 3, 4]) true iex&gt; Xema.validate(schema, [2, 3, 1]) {:error, %Xema.ValidationError{ reason: %{ items: [{2, %{value: 1, minimum: 2}}]} } } More examples can be found on page Usage."},{"ref":"Xema.html#source/1","title":"Xema.source/1","type":"function","doc":"Returns the source for a given xema. The output can differ from the input if the schema contains references. To get the original source the schema must be created with inline: false. Examples iex&gt; {:integer, minimum: 1} |&gt; Xema.new() |&gt; Xema.source() {:integer, minimum: 1}"},{"ref":"Xema.html#valid?/2","title":"Xema.valid?/2","type":"function","doc":"Returns true if the value is a valid value against the given schema; otherwise returns false."},{"ref":"Xema.html#validate/2","title":"Xema.validate/2","type":"function","doc":"Returns :ok if the value is a valid value against the given schema; otherwise returns an error tuple."},{"ref":"Xema.html#validate!/2","title":"Xema.validate!/2","type":"function","doc":"Returns :ok if the value is a valid value against the given schema; otherwise raises a Elixir.Xema.ValidationError."},{"ref":"Xema.html#t:t/0","title":"Xema.t/0","type":"type","doc":"This struct contains the schema and references of the schema."},{"ref":"Xema.Behaviour.html","title":"Xema.Behaviour","type":"behaviour","doc":"A behaviour module for implementing a schema validator. This behaviour is just for Xema and JsonXema."},{"ref":"Xema.Behaviour.html#c:init/1","title":"Xema.Behaviour.init/1","type":"callback","doc":"This callback initialize the schema. The function gets the data given to Xema.new/1 and returns a Xema.Schema."},{"ref":"Xema.Builder.html","title":"Xema.Builder","type":"module","doc":"This module contains some convenience functions to generate schemas. Examples iex&gt; import Xema.Builder ...&gt; schema = Xema.new integer(minimum: 1) ...&gt; Xema.valid?(schema, 6) true ...&gt; Xema.valid?(schema, 0) false"},{"ref":"Xema.Builder.html#any/0","title":"Xema.Builder.any/0","type":"function","doc":"Returns a tuple of :any and the given keyword list. Examples iex&gt; Xema.Builder.any(key: 42) {:any, [key: 42]}"},{"ref":"Xema.Builder.html#any/1","title":"Xema.Builder.any/1","type":"function","doc":""},{"ref":"Xema.Builder.html#atom/0","title":"Xema.Builder.atom/0","type":"function","doc":"Returns a tuple of :atom and the given keyword list. Examples iex&gt; Xema.Builder.atom(key: 42) {:atom, [key: 42]}"},{"ref":"Xema.Builder.html#atom/1","title":"Xema.Builder.atom/1","type":"function","doc":""},{"ref":"Xema.Builder.html#boolean/0","title":"Xema.Builder.boolean/0","type":"function","doc":"Returns a tuple of :boolean and the given keyword list. Examples iex&gt; Xema.Builder.boolean(key: 42) {:boolean, [key: 42]}"},{"ref":"Xema.Builder.html#boolean/1","title":"Xema.Builder.boolean/1","type":"function","doc":""},{"ref":"Xema.Builder.html#float/0","title":"Xema.Builder.float/0","type":"function","doc":"Returns a tuple of :float and the given keyword list. Examples iex&gt; Xema.Builder.float(key: 42) {:float, [key: 42]}"},{"ref":"Xema.Builder.html#float/1","title":"Xema.Builder.float/1","type":"function","doc":""},{"ref":"Xema.Builder.html#integer/0","title":"Xema.Builder.integer/0","type":"function","doc":"Returns a tuple of :integer and the given keyword list. Examples iex&gt; Xema.Builder.integer(key: 42) {:integer, [key: 42]}"},{"ref":"Xema.Builder.html#integer/1","title":"Xema.Builder.integer/1","type":"function","doc":""},{"ref":"Xema.Builder.html#keyword/0","title":"Xema.Builder.keyword/0","type":"function","doc":"Returns a tuple of :keyword and the given keyword list. Examples iex&gt; Xema.Builder.keyword(key: 42) {:keyword, [key: 42]}"},{"ref":"Xema.Builder.html#keyword/1","title":"Xema.Builder.keyword/1","type":"function","doc":""},{"ref":"Xema.Builder.html#list/0","title":"Xema.Builder.list/0","type":"function","doc":"Returns a tuple of :list and the given keyword list. Examples iex&gt; Xema.Builder.list(key: 42) {:list, [key: 42]}"},{"ref":"Xema.Builder.html#list/1","title":"Xema.Builder.list/1","type":"function","doc":""},{"ref":"Xema.Builder.html#map/0","title":"Xema.Builder.map/0","type":"function","doc":"Returns a tuple of :map and the given keyword list. Examples iex&gt; Xema.Builder.map(key: 42) {:map, [key: 42]}"},{"ref":"Xema.Builder.html#map/1","title":"Xema.Builder.map/1","type":"function","doc":""},{"ref":"Xema.Builder.html#number/0","title":"Xema.Builder.number/0","type":"function","doc":"Returns a tuple of :number and the given keyword list. Examples iex&gt; Xema.Builder.number(key: 42) {:number, [key: 42]}"},{"ref":"Xema.Builder.html#number/1","title":"Xema.Builder.number/1","type":"function","doc":""},{"ref":"Xema.Builder.html#ref/1","title":"Xema.Builder.ref/1","type":"function","doc":"Returns the tuple {:ref, ref}."},{"ref":"Xema.Builder.html#string/0","title":"Xema.Builder.string/0","type":"function","doc":"Returns a tuple of :string and the given keyword list. Examples iex&gt; Xema.Builder.string(key: 42) {:string, [key: 42]}"},{"ref":"Xema.Builder.html#string/1","title":"Xema.Builder.string/1","type":"function","doc":""},{"ref":"Xema.Builder.html#strux/0","title":"Xema.Builder.strux/0","type":"function","doc":"Returns :struct."},{"ref":"Xema.Builder.html#strux/1","title":"Xema.Builder.strux/1","type":"function","doc":"Returns the tuple {:struct, module: module}."},{"ref":"Xema.Builder.html#strux/2","title":"Xema.Builder.strux/2","type":"function","doc":""},{"ref":"Xema.Builder.html#tuple/0","title":"Xema.Builder.tuple/0","type":"function","doc":"Returns a tuple of :tuple and the given keyword list. Examples iex&gt; Xema.Builder.tuple(key: 42) {:tuple, [key: 42]}"},{"ref":"Xema.Builder.html#tuple/1","title":"Xema.Builder.tuple/1","type":"function","doc":""},{"ref":"Xema.Builder.html#xema/1","title":"Xema.Builder.xema/1","type":"macro","doc":"Creates a schema."},{"ref":"Xema.Builder.html#xema/2","title":"Xema.Builder.xema/2","type":"macro","doc":"Creates a schema with the given name."},{"ref":"Xema.CastError.html","title":"Xema.CastError","type":"exception","doc":"Raised when a cast fails."},{"ref":"Xema.CastError.html#format_error/1","title":"Xema.CastError.format_error/1","type":"function","doc":"Formats the error map to an error message."},{"ref":"Xema.CastError.html#t:error/0","title":"Xema.CastError.error/0","type":"type","doc":""},{"ref":"Xema.CastError.html#t:t/0","title":"Xema.CastError.t/0","type":"type","doc":""},{"ref":"Xema.Castable.html","title":"Xema.Castable","type":"protocol","doc":"Converts data using the specified schema."},{"ref":"Xema.Castable.html#cast/2","title":"Xema.Castable.cast/2","type":"function","doc":"Converts the given data using the specified schema."},{"ref":"Xema.Castable.html#t:t/0","title":"Xema.Castable.t/0","type":"type","doc":""},{"ref":"Xema.Caster.html","title":"Xema.Caster","type":"behaviour","doc":"A behaviour for a caster."},{"ref":"Xema.Caster.html#c:cast/1","title":"Xema.Caster.cast/1","type":"callback","doc":"A callback for a caster."},{"ref":"Xema.Format.html","title":"Xema.Format","type":"module","doc":"This module contains semantic validators for strings."},{"ref":"Xema.Format.html#date?/1","title":"Xema.Format.date?/1","type":"function","doc":"Checks if the string is a valid date representation. This function returns true if the value is a string and is formatted as defined by RFC 3339, false otherwise."},{"ref":"Xema.Format.html#date_time?/1","title":"Xema.Format.date_time?/1","type":"function","doc":"Checks if the string is a valid date time representation. This function returns true if the value is a string and is formatted as defined by RFC 3339, false otherwise."},{"ref":"Xema.Format.html#email?/1","title":"Xema.Format.email?/1","type":"function","doc":"Checks if the string is a valid email representation. This function returns true if the value is a string and is formatted as defined by RFC 5322, false otherwise. The regular expression was taken from https://emailregex.com/. Examples iex&gt; import Xema.Format iex&gt; iex&gt; email?(&quot;marin.musterman@germany.net&quot;) true iex&gt; email?(&quot;Otto.Normalverbraucher&quot;) false iex&gt; email?(&quot;Otto.Normal@Verbraucher.NET&quot;) true"},{"ref":"Xema.Format.html#host?/1","title":"Xema.Format.host?/1","type":"function","doc":"Checks if the string is a valid host representation. This function returns true if the value is a valid IPv4 address, IPv6 address, or a valid hostname, false otherwise. Examples iex&gt; import Xema.Format iex&gt; iex&gt; host?(&quot;127.0.0.1&quot;) true iex&gt; host?(&quot;localhost&quot;) true iex&gt; host?(&quot;elixirforum.com&quot;) true iex&gt; host?(&quot;go go go&quot;) false"},{"ref":"Xema.Format.html#hostname?/1","title":"Xema.Format.hostname?/1","type":"function","doc":"Checks if the string is a valid hostname representation. This function returns true if the value is a string and is formatted as defined by RFC 1034, false otherwise."},{"ref":"Xema.Format.html#ipv4?/1","title":"Xema.Format.ipv4?/1","type":"function","doc":"Checks if the string is a valid IPv4 address representation. This function returns true if the value is a string and is formatted as defined by RFC 2673, false otherwise."},{"ref":"Xema.Format.html#ipv6?/1","title":"Xema.Format.ipv6?/1","type":"function","doc":"Checks if the string is a valid IPv6 address representation. This function returns true if the value is a string and is formatted as defined by RFC 2373, false otherwise."},{"ref":"Xema.Format.html#is?/2","title":"Xema.Format.is?/2","type":"function","doc":"Checks if the value matches the given type. The function expected a available format and a string to check. Returns true for a valid string, false otherwise. Examples iex&gt; Xema.Format.is?(:email, &quot;foo@bar.net&quot;) true iex&gt; Xema.Format.is?(:email, &quot;foo.bar.net&quot;) false"},{"ref":"Xema.Format.html#json_pointer?/1","title":"Xema.Format.json_pointer?/1","type":"function","doc":"Checks if the string is a valid JSON pointer representation."},{"ref":"Xema.Format.html#regex?/1","title":"Xema.Format.regex?/1","type":"function","doc":"Return true if string contains a regular expression."},{"ref":"Xema.Format.html#relative_json_pointer?/1","title":"Xema.Format.relative_json_pointer?/1","type":"function","doc":"Checks if the string is a valid JSON pointer representation."},{"ref":"Xema.Format.html#supports/1","title":"Xema.Format.supports/1","type":"macro","doc":"Returns true for a valid format type, false otherwise. This macro can be used in guards."},{"ref":"Xema.Format.html#time?/1","title":"Xema.Format.time?/1","type":"function","doc":"Checks if the string is a valid time representation. This function returns true if the value is a string and is formatted as defined by RFC 3339, false otherwise."},{"ref":"Xema.Format.html#uri?/1","title":"Xema.Format.uri?/1","type":"function","doc":"Checks if the string is a valid URI representation. This function returns true if the value is a string and is formatted as defined by RFC 3986, false otherwise. The following are two example URIs and their component parts: hierarchical part | |-----------------------------------------| authority path | | |-------------------------------||--------| abc://username:password@example.com:123/path/data?key=value#fragid1 |-| |---------------| |---------| |-| |-------| |-----| | | | | | | scheme user information host port query fragment urn:example:mammal:monotreme:echidna |-| |------------------------------| | | scheme path Wikipedia: Uniform Resource Identifier"},{"ref":"Xema.Format.html#uri_fragment?/1","title":"Xema.Format.uri_fragment?/1","type":"function","doc":"Checks if the string is a valid URI fragment representation."},{"ref":"Xema.Format.html#uri_path?/1","title":"Xema.Format.uri_path?/1","type":"function","doc":"Checks if the string is a valid URI path representation. See also Xema.Format.uri?/1."},{"ref":"Xema.Format.html#uri_query?/1","title":"Xema.Format.uri_query?/1","type":"function","doc":"Checks if the string is a valid URI query representation."},{"ref":"Xema.Format.html#uri_reference?/1","title":"Xema.Format.uri_reference?/1","type":"function","doc":"Checks if the string is a valid URI reference representation."},{"ref":"Xema.Format.html#uri_template?/1","title":"Xema.Format.uri_template?/1","type":"function","doc":"Checks if the string is a valid URI template representation."},{"ref":"Xema.Format.html#uri_template_path?/1","title":"Xema.Format.uri_template_path?/1","type":"function","doc":"Checks if the string is a valid URI template path representation."},{"ref":"Xema.Format.html#uri_userinfo?/1","title":"Xema.Format.uri_userinfo?/1","type":"function","doc":"Checks if the string is a valid URI user info. See also Xema.Format.uri?/1."},{"ref":"Xema.Format.html#t:format/0","title":"Xema.Format.format/0","type":"type","doc":"The list of supported validators."},{"ref":"Xema.Loader.html","title":"Xema.Loader","type":"behaviour","doc":"The behaviour for loaders. For the loader configuration see &quot;Configure a loader&quot;."},{"ref":"Xema.Loader.html#fetch/1","title":"Xema.Loader.fetch/1","type":"function","doc":""},{"ref":"Xema.Loader.html#c:fetch/1","title":"Xema.Loader.fetch/1","type":"callback","doc":"This function expected an URI, to fetch the required data to create a schema."},{"ref":"Xema.NoLoader.html","title":"Xema.NoLoader","type":"module","doc":"The default loader. For the loader configuration see &quot;Configure a loader&quot;."},{"ref":"Xema.NoLoader.html#fetch/1","title":"Xema.NoLoader.fetch/1","type":"function","doc":"Returns for any data always the error tuple {:error, &quot;No loader configured.&quot;}."},{"ref":"Xema.Ref.html","title":"Xema.Ref","type":"module","doc":"This module contains a struct and functions to represent and handle references."},{"ref":"Xema.Ref.html#fetch!/3","title":"Xema.Ref.fetch!/3","type":"function","doc":"Returns the schema and the root for the given ref and xema."},{"ref":"Xema.Ref.html#fragment/1","title":"Xema.Ref.fragment/1","type":"function","doc":""},{"ref":"Xema.Ref.html#key/1","title":"Xema.Ref.key/1","type":"function","doc":"Returns the reference key for a Ref or an URI."},{"ref":"Xema.Ref.html#new/1","title":"Xema.Ref.new/1","type":"function","doc":"Creates a new reference from the given pointer."},{"ref":"Xema.Ref.html#new/2","title":"Xema.Ref.new/2","type":"function","doc":"Creates a new reference from the given pointer and uri."},{"ref":"Xema.Ref.html#validate/3","title":"Xema.Ref.validate/3","type":"function","doc":"Validates the given value with the referenced schema."},{"ref":"Xema.Ref.html#t:t/0","title":"Xema.Ref.t/0","type":"type","doc":"A reference contains a pointer and an optional uri."},{"ref":"Xema.Schema.html","title":"Xema.Schema","type":"module","doc":"This module contains the struct for the keywords of a schema."},{"ref":"Xema.Schema.html#fetch/2","title":"Xema.Schema.fetch/2","type":"function","doc":"Fetches a subschema from the schema by the given pointer. If schema contains the given pointer with a subschema, then {:ok, schema} is returned otherwise :error."},{"ref":"Xema.Schema.html#fetch!/2","title":"Xema.Schema.fetch!/2","type":"function","doc":"Fetches a subschema from the schema by the given pointer. If schema contains the given pointer with a subschema, then {:ok, schema} is returned otherwise a SchemaError is raised."},{"ref":"Xema.Schema.html#keywords/0","title":"Xema.Schema.keywords/0","type":"function","doc":"Returns all keywords in a list. The key :data is not a regular keyword and is not in the list."},{"ref":"Xema.Schema.html#new/1","title":"Xema.Schema.new/1","type":"function","doc":"Returns a %Schema{} for the given keywords in the keyword list."},{"ref":"Xema.Schema.html#to_map/1","title":"Xema.Schema.to_map/1","type":"function","doc":"Returns the %Schema{} as a map. Items which a nil value are not in the map."},{"ref":"Xema.Schema.html#types/0","title":"Xema.Schema.types/0","type":"function","doc":"Returns all available types in a list."},{"ref":"Xema.Schema.html#t:t/0","title":"Xema.Schema.t/0","type":"type","doc":"The struct contains the keywords for a schema. additional_items disallow additional items, if set to false. The keyword can also contain a schema to specify the type of additional items. additional_properties disallow additional properties, if set to true. &#39;all_of&#39; a list of schemas they must all be valid. &#39;any_of&#39; a list of schemas with any valid schema. caster a custom caster. This can be a function, a tuple with module and function name, or a Xema.Caster behaviour. comment for the schema. const specifies a constant. content_encoding annotation for the encoding. content_media_type annotation for the media type. contains validates a list whether any item is valid for the given schema. data none schema data. default this keyword can be used to supply a default JSON value. definitions contains schemas for reuse. dependencies allows the schema of the map to change based on the presence of certain special properties description of the schema. else see if, then, else. enum specifies an enumeration examples the value of this keyword must be an array. There are no restrictions placed on the values within the array. exclusive_maximum is a boolean. When true, it indicates that the range excludes the maximum value. exclusive_minimum is a boolean. When true, it indicates that the range excludes the minimum value. format semantic validation. id a unique identifier. if, then, else: These keywords work together to implement conditional application of a subschema based on the outcome of another subschema. items specifies the type(s) of the items. keys could be :atoms or :strings. max_items the maximum length of list. max_length the maximum length of string. max_properties the maximum count of properties for the map. maximum the maximum value. min_items the minimal length of list. min_length the minimal length of string. min_properties the minimal count of properties for the map. minimum the minimum value. module the module of a struct. multiple_of is a number greater 0. The value has to be a multiple of this number. not negates the given schema one_of the given data must be valid against exactly one of the given subschemas. pattern_properties specifies schemas for properties by patterns pattern restrict a string to a particular regular expression. properties specifies schemas for properties. property_names a schema to check property names. ref a reference to a schema. required contains a set of required properties. schema declares the used schema. title of the schema. then see if, then, else type specifies the data type for a schema. unique_items disallow duplicate items, if set to true. validator a custom validator. This can be a function, a tuple with module and function name, or a Xema.Validator behaviour."},{"ref":"Xema.Schema.html#t:type/0","title":"Xema.Schema.type/0","type":"type","doc":"The type for the schema."},{"ref":"Xema.SchemaError.html","title":"Xema.SchemaError","type":"exception","doc":"Raised when a schema can&#39;t be build."},{"ref":"Xema.SchemaValidator.html","title":"Xema.SchemaValidator","type":"module","doc":"A validator for the schema input."},{"ref":"Xema.SchemaValidator.html#validate!/1","title":"Xema.SchemaValidator.validate!/1","type":"function","doc":"Returns :ok if the data valid against the xema schema."},{"ref":"Xema.Utils.html","title":"Xema.Utils","type":"module","doc":"Some utilities for Xema."},{"ref":"Xema.Utils.html#has_key?/2","title":"Xema.Utils.has_key?/2","type":"function","doc":"Returns whether the given key exists in the given value. Returns true if value is a map and contains key as a key. value is a keyword and contains key as a key. value is a list of tuples with keyas the first element. Example iex&gt; alias Xema.Utils iex&gt; Utils.has_key?(%{foo: 5}, :foo) true iex&gt; Utils.has_key?([foo: 5], :foo) true iex&gt; Utils.has_key?([{&quot;foo&quot;, 5}], &quot;foo&quot;) true"},{"ref":"Xema.Utils.html#size/1","title":"Xema.Utils.size/1","type":"function","doc":"Returns the size of a list or tuple."},{"ref":"Xema.Utils.html#to_existing_atom/1","title":"Xema.Utils.to_existing_atom/1","type":"function","doc":"Converts the given string to an existing atom. Returns nil if the atom does not exist. Examples iex&gt; import Xema.Utils iex&gt; to_existing_atom(:my_atom) :my_atom iex&gt; to_existing_atom(&quot;my_atom&quot;) :my_atom iex&gt; to_existing_atom(&quot;not_existing_atom&quot;) nil"},{"ref":"Xema.Utils.html#update_uri/2","title":"Xema.Utils.update_uri/2","type":"function","doc":"Returns nil if uri_1 and uri_2 are nil. Parses a URI when the other URI is nil. Merges URIs if both are not nil."},{"ref":"Xema.ValidationError.html","title":"Xema.ValidationError","type":"exception","doc":"Raised when a validation fails."},{"ref":"Xema.ValidationError.html#format_error/1","title":"Xema.ValidationError.format_error/1","type":"function","doc":"This function returns an error message for an error or error tuple. Example iex&gt; schema = Xema.new(:integer) iex&gt; schema ...&gt; |&gt; Xema.Validator.validate(1.1) ...&gt; |&gt; Xema.ValidationError.format_error() &quot;Expected :integer, got 1.1.&quot;"},{"ref":"Xema.ValidationError.html#travers_errors/4","title":"Xema.ValidationError.travers_errors/4","type":"function","doc":"Traverse the error tree and invokes the given function. Example iex&gt; fun = fn _error, path, acc -&gt; ...&gt; [&quot;Error at &quot; &lt;&gt; inspect(path) | acc] ...&gt; end iex&gt; iex&gt; schema = Xema.new( ...&gt; properties: %{ ...&gt; int: :integer, ...&gt; names: {:list, items: :string}, ...&gt; num: [any_of: [:integer, :float]] ...&gt; } ...&gt; ) iex&gt; iex&gt; data = %{int: &quot;x&quot;, names: [1, &quot;x&quot;, 5], num: :foo} iex&gt; iex&gt; schema ...&gt; |&gt; Xema.Validator.validate(data) ...&gt; |&gt; Xema.ValidationError.travers_errors([], fun) [ &quot;Error at [:num]&quot;, &quot;Error at [:names, 2]&quot;, &quot;Error at [:names, 0]&quot;, &quot;Error at [:names]&quot;, &quot;Error at [:int]&quot;, &quot;Error at []&quot; ]"},{"ref":"Xema.ValidationError.html#t:opts/0","title":"Xema.ValidationError.opts/0","type":"type","doc":""},{"ref":"Xema.ValidationError.html#t:path/0","title":"Xema.ValidationError.path/0","type":"type","doc":""},{"ref":"Xema.Validator.html","title":"Xema.Validator","type":"behaviour","doc":"This module contains all validators to check data against a schema."},{"ref":"Xema.Validator.html#c:validate/1","title":"Xema.Validator.validate/1","type":"callback","doc":"A callback for custom validators. For an example see: Custom validators"},{"ref":"Xema.Validator.html#validate/2","title":"Xema.Validator.validate/2","type":"function","doc":"Validates data against the given schema."},{"ref":"Xema.Validator.html#t:result/0","title":"Xema.Validator.result/0","type":"type","doc":""},{"ref":"readme.html","title":"Xema","type":"extras","doc":"Xema Xema is a schema validator inspired by JSON Schema. Xema supports the features documented in draft 04, 06, and 07 of the JSON-Schema specification. Xema allows you to annotate and validate elixir data structures. Xema is in early beta. If you try it and has an issue, report them."},{"ref":"usage.html","title":"Usage","type":"extras","doc":"Usage This page describes all keywords that are available for creating a schema. All schemas will construct in the &quot;raw&quot; format. The Xema.Builder provides some convenience functions to create schemas. The recommended method to construct a schema is use Xema described in the module documentation of Xema."},{"ref":"usage.html#type-any","title":"Usage - Type any","type":"extras","doc":"The schema any will accept any data. iex&gt; schema = Xema.new :any iex&gt; Xema.validate schema, 42 :ok iex&gt; Xema.validate schema, &quot;foo&quot; :ok iex&gt; Xema.validate schema, nil :ok"},{"ref":"usage.html#type-nil","title":"Usage -  Type nil","type":"extras","doc":"The nil type matches only nil. iex&gt; schema = Xema.new :nil iex&gt; Xema.validate schema, nil :ok iex&gt; {:error, error} = Xema.validate schema, 0 {:error, %Xema.ValidationError{ reason: %{type: :nil, value: 0} }} iex&gt; Exception.message(error) &quot;Expected nil, got 0.&quot;"},{"ref":"usage.html#type-boolean","title":"Usage -  Type boolean","type":"extras","doc":"The boolean type matches only true and false. iex&gt; schema = Xema.new :boolean iex&gt; Xema.validate schema, true :ok iex&gt; Xema.valid? schema, false true iex&gt; {:error, error} = Xema.validate schema, 0 {:error, %Xema.ValidationError{ reason: %{type: :boolean, value: 0} }} iex&gt; Exception.message(error) &quot;Expected :boolean, got 0.&quot; iex&gt; Xema.valid? schema, nil false"},{"ref":"usage.html#type-atom","title":"Usage -  Type atom","type":"extras","doc":"The atom type matches only atoms. Schemas of type atom match also the atoms true, false, and nil. iex&gt; schema = Xema.new :atom iex&gt; Xema.validate schema, :foo :ok iex&gt; Xema.valid? schema, &quot;foo&quot; false iex&gt; {:error, error} = Xema.validate schema, 0 {:error, %Xema.ValidationError{ reason: %{type: :atom, value: 0} }} iex&gt; Exception.message(error) &quot;Expected :atom, got 0.&quot; iex&gt; Xema.valid? schema, nil true iex&gt; Xema.valid? schema, false true"},{"ref":"usage.html#type-string","title":"Usage -  Type string","type":"extras","doc":"JSON Schema Draft: 4/6/7 The string type is used for strings. iex&gt; schema = Xema.new :string iex&gt; Xema.validate schema, &quot;José&quot; :ok iex&gt; {:error, error} = Xema.validate schema, 42 {:error, %Xema.ValidationError{ reason: %{type: :string, value: 42} }} iex&gt; Exception.message(error) &quot;Expected :string, got 42.&quot; iex&gt; Xema.valid? schema, &quot;José&quot; true iex&gt; Xema.valid? schema, 42 false Length The length of a string can be constrained using the min_length and max_length keywords. For both keywords, the value must be a non-negative number. iex&gt; schema = Xema.new {:string, min_length: 2, max_length: 3} iex&gt; {:error, error} = Xema.validate schema, &quot;a&quot; {:error, %Xema.ValidationError{ reason: %{value: &quot;a&quot;, min_length: 2} }} iex&gt; Exception.message(error) ~s|Expected minimum length of 2, got &quot;a&quot;.| iex&gt; Xema.validate schema, &quot;ab&quot; :ok iex&gt; Xema.validate schema, &quot;abc&quot; :ok iex&gt; {:error, error} = Xema.validate schema, &quot;abcd&quot; {:error, %Xema.ValidationError{ reason: %{value: &quot;abcd&quot;, max_length: 3} }} iex&gt; Exception.message(error) ~s|Expected maximum length of 3, got &quot;abcd&quot;.| Regular Expression The pattern keyword is used to restrict a string to a particular regular expression. iex&gt; schema = Xema.new {:string, pattern: ~r/[0-9]-[A-B]+/} iex&gt; Xema.validate schema, &quot;1-AB&quot; :ok iex&gt; {:error, error} = Xema.validate schema, &quot;foo&quot; {:error, %Xema.ValidationError{ reason: %{value: &quot;foo&quot;, pattern: ~r/[0-9]-[A-B]+/} }} iex&gt; Exception.message(error) ~s|Pattern ~r/[0-9]-[A-B]+/ does not match value &quot;foo&quot;.| The regular expression can also be a string. iex&gt; schema = Xema.new {:string, pattern: &quot;[0-9]-[A-B]+&quot;} iex&gt; Xema.validate schema, &quot;1-AB&quot; :ok iex&gt; {:error, error} = Xema.validate schema, &quot;foo&quot; {:error, %Xema.ValidationError{ reason: %{value: &quot;foo&quot;, pattern: ~r/[0-9]-[A-B]+/} }} iex&gt; Exception.message(error) ~s|Pattern ~r/[0-9]-[A-B]+/ does not match value &quot;foo&quot;.| Format JSON Schema Draft: 4/6/7 Basic semantic validation of strings. :date_time validation as defined by RFC 3339 iex&gt; schema = Xema.new {:string, format: :date_time} iex&gt; Xema.valid? schema, &quot;today&quot; false iex&gt; Xema.valid? schema, &quot;1963-06-19T08:30:06.283185Z&quot; true :email validation as defined by RFC 5322 iex&gt; {:string, format: :email} ...&gt; |&gt; Xema.new() ...&gt; |&gt; Xema.valid?(&quot;marion.mustermann@otto.net&quot;) true :host checks if the string is an valid IPv4, IPv6, or hostname. :hostname validation as defined by RFC 1034 :ipv4 validation as defined by RFC 2673 :ipv6 validation as defined by RFC 2373 :uri validation as defindex by RFC 3986 :uri_fragment :uri_path :uri_query :uri_userinfo JSON Schema Draft: -/-/7 :regexchecks if the string is a valid regular expression."},{"ref":"usage.html#types-number-integer-and-float","title":"Usage -  Types number, integer and float","type":"extras","doc":"There are three numeric types in Xema: number, integer and float. They share the same validation keywords. The number type is used for numbers. iex&gt; schema = Xema.new :number iex&gt; Xema.validate schema, 42 :ok iex&gt; Xema.validate schema, 21.5 :ok iex&gt; {:error, error} = Xema.validate schema, &quot;foo&quot; {:error, %Xema.ValidationError{ reason: %{type: :number, value: &quot;foo&quot;} }} iex&gt; Exception.message(error) ~s|Expected :number, got &quot;foo&quot;.| The integer type is used for integral numbers. iex&gt; schema = Xema.new :integer iex&gt; Xema.validate schema, 42 :ok iex&gt; {:error, error} = Xema.validate schema, 21.5 {:error, %Xema.ValidationError{ reason: %{type: :integer, value: 21.5} }} iex&gt; Exception.message(error) &quot;Expected :integer, got 21.5.&quot; The float type is used for floating point numbers. iex&gt; schema = Xema.new :float iex&gt; {:error, error} = Xema.validate schema, 42 {:error, %Xema.ValidationError{ reason: %{type: :float, value: 42} }} iex&gt; Exception.message(error) &quot;Expected :float, got 42.&quot; iex&gt; Xema.validate schema, 21.5 :ok Multiples JSON Schema Draft: 4/6/7 Numbers can be restricted to a multiple of a given number, using the multiple_of keyword. It may be set to any positive number. iex&gt; schema = Xema.new {:number, multiple_of: 2} iex&gt; Xema.validate schema, 8 :ok iex&gt; {:error, error} = Xema.validate schema, 7 {:error, %Xema.ValidationError{ reason: %{value: 7, multiple_of: 2} }} iex&gt; Exception.message(error) &quot;Value 7 is not a multiple of 2.&quot; iex&gt; Xema.valid? schema, 8.0 true Range JSON Schema Draft: 4/-/- Ranges of numbers are specified using a combination of the minimum, maximum, exclusive_minimum and exclusive_maximum keywords. minimum specifies a minimum numeric value. exclusive_minimum is a boolean. When true, it indicates that the range excludes the minimum value, i.e., x &gt; minx &gt; min. When false (or not included), it indicates that the range includes the minimum value, i.e., x≥minx≥min. maximum specifies a maximum numeric value. exclusive_maximum is a boolean. When true, it indicates that the range excludes the maximum value, i.e., x &lt; maxx &lt; max. When false (or not included), it indicates that the range includes the maximum value, i.e., x ≤ maxx ≤ max. iex&gt; schema = Xema.new { ...&gt; :float, ...&gt; minimum: 1.2, maximum: 1.4, exclusive_maximum: true ...&gt; } iex&gt; {:error, error} = Xema.validate schema, 1.1 {:error, %Xema.ValidationError{ reason: %{value: 1.1, minimum: 1.2} }} iex&gt; Exception.message(error) &quot;Value 1.1 is less than minimum value of 1.2.&quot; iex&gt; Xema.validate schema, 1.2 :ok iex&gt; Xema.valid? schema, 1.3 true iex&gt; {:error, error} = Xema.validate schema, 1.4 {:error, %Xema.ValidationError{ reason: %{value: 1.4, maximum: 1.4, exclusive_maximum: true} }} iex&gt; Exception.message(error) &quot;Value 1.4 equals exclusive maximum value of 1.4.&quot; iex&gt; {:error, error} = Xema.validate schema, 1.5 {:error, %Xema.ValidationError{ reason: %{value: 1.5, maximum: 1.4, exclusive_maximum: true} }} iex&gt; Exception.message(error) &quot;Value 1.5 exceeds maximum value of 1.4.&quot; JSON Schema Draft: 6/7 The keywords exclusive_maximum and exclusive_minimum changed from a boolean to a number. Wherever one of these would be true before, they have now the value of the corresponding keyword maximum or minimum. The keyword maximum/minimum can be removed. iex&gt; schema = Xema.new {:float, minimum: 1.2, exclusive_maximum: 1.4} iex&gt; {:error, error} = Xema.validate schema, 1.1 {:error, %Xema.ValidationError{ reason: %{value: 1.1, minimum: 1.2} }} iex&gt; Exception.message(error) &quot;Value 1.1 is less than minimum value of 1.2.&quot; iex&gt; Xema.validate schema, 1.2 :ok iex&gt; Xema.valid? schema, 1.3 true iex&gt; {:error, error} = Xema.validate schema, 1.4 {:error, %Xema.ValidationError{ reason: %{value: 1.4, exclusive_maximum: 1.4} }} iex&gt; Exception.message(error) &quot;Value 1.4 equals exclusive maximum value of 1.4.&quot; iex&gt; {:error, error} = Xema.validate schema, 1.5 {:error, %Xema.ValidationError{ reason: %{value: 1.5, exclusive_maximum: 1.4} }} iex&gt; Exception.message(error) &quot;Value 1.5 exceeds maximum value of 1.4.&quot;"},{"ref":"usage.html#type-list","title":"Usage -  Type list","type":"extras","doc":"List are used for ordered elements, each element may be of a different type. iex&gt; schema = Xema.new :list iex&gt; Xema.valid? schema, [1, &quot;two&quot;, 3.0] true iex&gt; {:error, error} = Xema.validate schema, 9 {:error, %Xema.ValidationError{ reason: %{type: :list, value: 9} }} iex&gt; Exception.message(error) &quot;Expected :list, got 9.&quot; Items The items keyword will be used to validate all items of a list to a single schema. iex&gt; schema = Xema.new {:list, items: :string} iex&gt; Xema.valid? schema, [&quot;a&quot;, &quot;b&quot;, &quot;abc&quot;] true iex&gt; {:error, error} = Xema.validate schema, [&quot;a&quot;, 1] {:error, %Xema.ValidationError{ reason: %{items: [{1, %{type: :string, value: 1}}]} }} iex&gt; Exception.message(error) &quot;Expected :string, got 1, at [1].&quot; The next example shows how to add keywords to the items schema. iex&gt; schema = Xema.new {:list, items: {:integer, minimum: 1, maximum: 10}} iex&gt; Xema.validate schema, [1, 2, 3] :ok iex&gt; {:error, error} = Xema.validate schema, [3, 2, 1, 0] {:error, %Xema.ValidationError{ reason: %{items: [{3, %{value: 0, minimum: 1}}]} }} iex&gt; Exception.message(error) &quot;Value 0 is less than minimum value of 1, at [3].&quot; items can also be used to give each item a specific schema. iex&gt; schema = Xema.new { ...&gt; :list, ...&gt; items: [:integer, {:string, min_length: 5}] ...&gt; } iex&gt; Xema.valid? schema, [1, &quot;hello&quot;] true iex&gt; {:error, error} = Xema.validate schema, [1, &quot;five&quot;] {:error, %Xema.ValidationError{ reason: %{items: [{1, %{value: &quot;five&quot;, min_length: 5}}]} }} iex&gt; Exception.message(error) ~s|Expected minimum length of 5, got &quot;five&quot;, at [1].| # It’s okay to not provide all of the items: iex&gt; Xema.validate schema, [1] :ok # And, by default, it’s also okay to add additional items to end: iex&gt; Xema.validate schema, [1, &quot;hello&quot;, &quot;foo&quot;] :ok Additional Items The additional_items keyword controls whether it is valid to have additional items in the array beyond what is defined in the schema. iex&gt; schema = Xema.new { ...&gt; :list, ...&gt; items: [:integer, {:string, min_length: 5}], ...&gt; additional_items: false ...&gt; } iex&gt; Xema.validate schema, [1] :ok # It’s okay to not provide all of the items: # But, since additionalItems is false, we can’t provide extra items: iex&gt; {:error, error} = Xema.validate schema, [1, &quot;hello&quot;, &quot;foo&quot;] {:error, %Xema.ValidationError{ reason: %{items: [{2, %{additional_items: false}}]} }} iex&gt; Exception.message(error) &quot;Unexpected additional item, at [2].&quot; iex&gt; {:error, error} = Xema.validate schema, [1, &quot;hello&quot;, &quot;foo&quot;, &quot;bar&quot;] {:error, %Xema.ValidationError{ reason: %{ items: [ {2, %{additional_items: false}}, {3, %{additional_items: false}} ] } }} iex&gt; Exception.message(error) &quot;&quot;&quot; Unexpected additional item, at [2]. Unexpected additional item, at [3].\\ &quot;&quot;&quot; The keyword can also contain a schema to specify the type of additional items. iex&gt; schema = Xema.new { ...&gt; :list, ...&gt; items: [:integer, {:string, min_length: 3}], ...&gt; additional_items: :integer ...&gt; } iex&gt; Xema.valid? schema, [1, &quot;two&quot;, 3, 4] true iex&gt; {:error, error} = Xema.validate schema, [1, &quot;two&quot;, 3, &quot;four&quot;] {:error, %Xema.ValidationError{ reason: %{items: [{3, %{type: :integer, value: &quot;four&quot;}}]} }} iex&gt; Exception.message(error) ~s|Expected :integer, got &quot;four&quot;, at [3].| Length The length of the array can be specified using the min_items and max_items keywords. The value of each keyword must be a non-negative number. iex&gt; schema = Xema.new {:list, min_items: 2, max_items: 3} iex&gt; {:error, error} = Xema.validate schema, [1] {:error, %Xema.ValidationError{ reason: %{value: [1], min_items: 2} }} iex&gt; Exception.message(error) &quot;Expected at least 2 items, got [1].&quot; iex&gt; Xema.validate schema, [1, 2] :ok iex&gt; Xema.validate schema, [1, 2, 3] :ok iex&gt; {:error, error} = Xema.validate schema, [1, 2, 3, 4] {:error, %Xema.ValidationError{ reason: %{value: [1, 2, 3, 4], max_items: 3} }} iex&gt; Exception.message(error) &quot;Expected at most 3 items, got [1, 2, 3, 4].&quot; Uniqueness A schema can ensure that each of the items in an array is unique. iex&gt; schema = Xema.new {:list, unique_items: true} iex&gt; Xema.valid? schema, [1, 2, 3] true iex&gt; {:error, error} = Xema.validate schema, [1, 2, 3, 2, 1] {:error, %Xema.ValidationError{ reason: %{value: [1, 2, 3, 2, 1], unique_items: true} }} iex&gt; Exception.message(error) &quot;Expected unique items, got [1, 2, 3, 2, 1].&quot;"},{"ref":"usage.html#type-tuple","title":"Usage -  Type tuple","type":"extras","doc":"Tuples are intended as fixed-size containers for multiple elements. The validation of tuples is similar to lists. iex&gt; schema = Xema.new {:tuple, min_items: 2, max_items: 3} iex&gt; {:error, error} = Xema.validate schema, {1} {:error, %Xema.ValidationError{ reason: %{value: {1}, min_items: 2} }} iex&gt; Exception.message(error) &quot;Expected at least 2 items, got {1}.&quot; iex&gt; Xema.validate schema, {1, 2} :ok iex&gt; Xema.validate schema, {1, 2, 3} :ok iex&gt; {:error, error} = Xema.validate schema, {1, 2, 3, 4} {:error, %Xema.ValidationError{ reason: %{value: {1, 2, 3, 4}, max_items: 3} }} iex&gt; Exception.message(error) &quot;Expected at most 3 items, got {1, 2, 3, 4}.&quot;"},{"ref":"usage.html#type-map","title":"Usage -  Type map","type":"extras","doc":"Whenever you need a key-value store, maps are the “go to” data structure in Elixir. Each of these pairs is conventionally referred to as a “property”. iex&gt; schema = Xema.new :map iex&gt; Xema.valid? schema, %{&quot;foo&quot; =&gt; &quot;bar&quot;} true iex&gt; {:error, error} = Xema.validate schema, &quot;bar&quot; {:error, %Xema.ValidationError{ reason: %{type: :map, value: &quot;bar&quot;} }} iex&gt; Exception.message(error) ~s|Expected :map, got &quot;bar&quot;.| # Using non-strings as keys are also valid: iex&gt; Xema.valid? schema, %{foo: &quot;bar&quot;} true iex&gt; Xema.valid? schema, %{1 =&gt; &quot;bar&quot;} true Keys The keyword keys can restrict the keys to atoms or strings. Atoms as keys: iex&gt; schema = Xema.new {:map, keys: :atoms} iex&gt; Xema.valid? schema, %{&quot;foo&quot; =&gt; &quot;bar&quot;} false iex&gt; Xema.valid? schema, %{foo: &quot;bar&quot;} true iex&gt; Xema.valid? schema, %{1 =&gt; &quot;bar&quot;} false Strings as keys: iex&gt; schema = Xema.new {:map, keys: :strings} iex&gt; Xema.valid? schema, %{&quot;foo&quot; =&gt; &quot;bar&quot;} true iex&gt; Xema.valid? schema, %{foo: &quot;bar&quot;} false iex&gt; Xema.valid? schema, %{1 =&gt; &quot;bar&quot;} false Properties The properties on a map are defined using the properties keyword. The value of properties is a map, where each key is the name of a property and each value is a schema used to validate that property. iex&gt; schema = Xema.new {:map, ...&gt; properties: %{ ...&gt; a: :integer, ...&gt; b: {:string, min_length: 5} ...&gt; } ...&gt; } iex&gt; Xema.valid? schema, %{a: 5, b: &quot;hello&quot;} true iex&gt; {:error, error} = Xema.validate schema, %{a: 5, b: &quot;ups&quot;} {:error, %Xema.ValidationError{ reason: %{ properties: %{ b: %{value: &quot;ups&quot;, min_length: 5} } } }} iex&gt; Exception.message(error) &quot;Expected minimum length of 5, got \\&quot;ups\\&quot;, at [:b].&quot; # Additinonal properties are allowed by default: iex&gt; Xema.valid? schema, %{a: 5, b: &quot;hello&quot;, add: :prop} true Required Properties By default, the properties defined by the properties keyword are not required. However, one can provide a list of required properties using the required keyword. iex&gt; schema = Xema.new {:map, properties: %{foo: :string}, required: [:foo]} iex&gt; Xema.validate schema, %{foo: &quot;bar&quot;} :ok iex&gt; {:error, error} = Xema.validate schema, %{bar: &quot;foo&quot;} {:error, %Xema.ValidationError{ reason: %{required: [:foo]} }} iex&gt; Exception.message(error) &quot;Required properties are missing: [:foo].&quot; Additional Properties The additional_properties keyword is used to control the handling of extra stuff, that is, properties whose names are not listed in the properties keyword. By default any additional properties are allowed. The additional_properties keyword may be either a boolean or an schema. If additional_properties is a boolean and set to false, no additional properties will be allowed. iex&gt; schema = Xema.new {:map, ...&gt; properties: %{foo: :string}, ...&gt; required: [:foo], ...&gt; additional_properties: false ...&gt; } iex&gt; Xema.validate schema, %{foo: &quot;bar&quot;} :ok iex&gt; {:error, error} = Xema.validate schema, %{foo: &quot;bar&quot;, bar: &quot;foo&quot;} {:error, %Xema.ValidationError{ reason: %{ properties: %{bar: %{additional_properties: false}} } }} iex&gt; Exception.message(error) &quot;Expected only defined properties, got key [:bar].&quot; additional_properties can also contain a schema to specify the type of additional properites. iex&gt; schema = Xema.new { ...&gt; :map, ...&gt; properties: %{foo: :string}, ...&gt; additional_properties: :integer ...&gt; } iex&gt; Xema.valid? schema, %{foo: &quot;foo&quot;, add: 1} true iex&gt; {:error, error} = Xema.validate schema, %{foo: &quot;foo&quot;, add: &quot;one&quot;} {:error, %Xema.ValidationError{ reason: %{properties: %{add: %{type: :integer, value: &quot;one&quot;}}} }} iex&gt; Exception.message(error) ~s|Expected :integer, got &quot;one&quot;, at [:add].| Pattern Properties The keyword pattern_properties defined additional properties by regular expressions. iex&gt; schema = Xema.new { ...&gt; :map, ...&gt; additional_properties: false, ...&gt; pattern_properties: %{ ...&gt; ~r/^s_/ =&gt; :string, ...&gt; ~r/^i_/ =&gt; :integer ...&gt; } ...&gt; } iex&gt; Xema.valid? schema, %{&quot;s_0&quot; =&gt; &quot;foo&quot;, &quot;i_1&quot; =&gt; 6} true iex&gt; Xema.valid? schema, %{s_0: &quot;foo&quot;, i_1: 6} true iex&gt; {:error, error} = Xema.validate schema, %{s_0: &quot;foo&quot;, f_1: 6.6} {:error, %Xema.ValidationError{ reason: %{properties: %{f_1: %{additional_properties: false}}} }} iex&gt; Exception.message(error) &quot;Expected only defined properties, got key [:f_1].&quot; Size The number of properties on an object can be restricted using the min_properties and max_properties keywords. iex&gt; schema = Xema.new {:map, ...&gt; min_properties: 2, ...&gt; max_properties: 3 ...&gt; } iex&gt; Xema.valid? schema, %{a: 1, b: 2} true iex&gt; {:error, error} = Xema.validate schema, %{} {:error, %Xema.ValidationError{ reason: %{min_properties: 2, value: %{}} }} iex&gt; Exception.message(error) &quot;Expected at least 2 properties, got %{}.&quot; iex&gt; {:error, error} = Xema.validate schema, %{a: 1, b: 2, c: 3, d: 4} {:error, %Xema.ValidationError{ reason: %{max_properties: 3, value: %{a: 1, b: 2, c: 3, d: 4}} }} iex&gt; Exception.message(error) &quot;Expected at most 3 properties, got %{a: 1, b: 2, c: 3, d: 4}.&quot; Size The type of a key in the schema also matters in validation. iex&gt; schema = Xema.new {:map, ...&gt; properties: %{foo: :integer}, ...&gt; additional_properties: false ...&gt; } iex&gt; Xema.valid? schema, %{foo: 1} true iex&gt; {:error, error} = Xema.validate schema, %{&quot;foo&quot; =&gt; 1} {:error, %Xema.ValidationError{ reason: %{properties: %{&quot;foo&quot; =&gt; %{additional_properties: false}}} }} iex&gt; Exception.message(error) ~s|Expected only defined properties, got key [\\&quot;foo\\&quot;].| Dependencies The dependencies keyword allows the schema of the object to change based on the presence of certain special properties. iex&gt; schema = Xema.new { ...&gt; :map, ...&gt; properties: %{ ...&gt; a: :number, ...&gt; b: :number, ...&gt; c: :number ...&gt; }, ...&gt; dependencies: %{ ...&gt; b: [:c] ...&gt; } ...&gt; } iex&gt; Xema.valid? schema, %{a: 5} true iex&gt; Xema.valid? schema, %{c: 9} true iex&gt; Xema.valid? schema, %{b: 1} false iex&gt; Xema.valid? schema, %{b: 1, c: 7} true"},{"ref":"usage.html#type-struct","title":"Usage -  Type struct","type":"extras","doc":"Structs can also be validated. iex&gt; schema = Xema.new :struct iex&gt; Xema.valid? schema, ~r/.*/ true iex&gt; Xema.valid? schema, %{} false Module The module keyword allows specifing which struct is expected. iex&gt; schema = Xema.new {:struct, module: Regex} iex&gt; Xema.valid? schema, ~r/.*/ true iex&gt; Xema.valid? schema, URI.parse(&quot;&quot;) false Map keywords for structs The validations for map are also available. The next example shows a schema for an URI that needs a fragment. iex&gt; schema = Xema.new { ...&gt; :struct, module: URI, properties: %{fragment: :string} ...&gt; } iex&gt; iex&gt; Xema.valid?(schema, URI.parse(&quot;http://example.com&quot;)) false iex&gt; Xema.valid?(schema, URI.parse(&quot;http://example.com#frag&quot;)) true"},{"ref":"usage.html#multiples-types","title":"Usage -  Multiples Types","type":"extras","doc":"JSON Schema Draft: 4/6/7 It is also possible to check if a value matches one of several types. iex&gt; schema = Xema.new {[:string, nil], min_length: 1} iex&gt; Xema.valid? schema, &quot;foo&quot; true iex&gt; Xema.valid? schema, nil true iex&gt; Xema.valid? schema, &quot;&quot; false"},{"ref":"usage.html#allow-additional-types","title":"Usage -  Allow Additional Types","type":"extras","doc":"JSON Schema Draft: - The keyword allow adds an extra type to the schema validation. iex&gt; schema = Xema.new {:string, min_length: 1, allow: nil} iex&gt; Xema.valid? schema, &quot;foo&quot; true iex&gt; Xema.valid? schema, nil true iex&gt; Xema.valid? schema, &quot;&quot; false"},{"ref":"usage.html#constants","title":"Usage -  Constants","type":"extras","doc":"JSON Schema Draft: -/6/7 This keyword checks if a value is equals to the given const. iex&gt; schema = Xema.new(const: 4711) iex&gt; Xema.validate schema, 4711 :ok iex&gt; {:error, error} = Xema.validate schema, 333 {:error, %Xema.ValidationError{ reason: %{const: 4711, value: 333} }} iex&gt; Exception.message(error) &quot;Expected 4711, got 333.&quot;"},{"ref":"usage.html#enumerations","title":"Usage -  Enumerations","type":"extras","doc":"The enum keyword is used to restrict a value to a fixed set of values. It must be an array with at least one element, where each element is unique. iex&gt; schema = Xema.new {:any, enum: [1, &quot;foo&quot;, :bar]} iex&gt; Xema.valid? schema, :bar true iex&gt; Xema.valid? schema, 42 false"},{"ref":"usage.html#negate-schema","title":"Usage -  Negate Schema","type":"extras","doc":"The keyword not negates a schema. iex&gt; schema = Xema.new(not: {:integer, minimum: 0}) iex&gt; Xema.valid? schema, 10 false iex&gt; Xema.valid? schema, -10 true"},{"ref":"usage.html#if-then-else","title":"Usage -  If-Then-Else","type":"extras","doc":"The keywords if, then, else work together to implement conditional application of a subschema based on the outcome of another subschema. iex&gt; schema = ...&gt; Xema.new( ...&gt; if: :list, ...&gt; then: [items: :integer, min_items: 2], ...&gt; else: :integer ...&gt; ) ...&gt; ...&gt; Xema.valid?(schema, 3) true iex&gt; Xema.valid?(schema, &quot;3&quot;) false iex&gt; Xema.valid?(schema, [1]) false iex&gt; Xema.valid?(schema, [1, 2]) true"},{"ref":"usage.html#custom-validator","title":"Usage -  Custom validator","type":"extras","doc":"With the keyword validator a custom validator can be defined. The validator expected a function or a tuple of module and function name. The validator function gets the current value and return :ok on success and an error tuple on failure. iex&gt; defmodule Palindrome do ...&gt; @xema Xema.new( ...&gt; properties: %{ ...&gt; palindrome: {:string, validator: &amp;Palindrome.check/1} ...&gt; } ...&gt; ) ...&gt; ...&gt; def check(value) do ...&gt; case value == String.reverse(value) do ...&gt; true -&gt; :ok ...&gt; false -&gt; {:error, :no_palindrome} ...&gt; end ...&gt; end ...&gt; ...&gt; def validate(value) do ...&gt; Xema.validate(@xema, value) ...&gt; end ...&gt; end iex&gt; iex&gt; Palindrome.validate(%{palindrome: &quot;abba&quot;}) :ok iex&gt; {:error, error} = Palindrome.validate(%{palindrome: &quot;beatles&quot;}) {:error, %Xema.ValidationError{ reason: %{ properties: %{ palindrome: %{validator: :no_palindrome, value: &quot;beatles&quot;} } } }} iex&gt; Exception.message(error) ~s|Validator fails with :no_palindrome for value &quot;beatles&quot;, at [:palindrome].|"},{"ref":"usage.html#combine-schemas","title":"Usage -  Combine Schemas","type":"extras","doc":"The keywords all_of, any_of, and one_of combines schemas. With all_of all schemas have to match. iex&gt; all = Xema.new(all_of: [ ...&gt; {:integer, multiple_of: 2}, ...&gt; {:integer, multiple_of: 3} ...&gt; ]) iex&gt; 0..9 |&gt; Enum.map(&amp;Xema.valid?(all, &amp;1)) |&gt; Enum.with_index() [true: 0, false: 1, false: 2, false: 3, false: 4, false: 5, true: 6, false: 7, false: 8, false: 9] With any_of any schema have to match. iex&gt; any = Xema.new(any_of: [ ...&gt; {:integer, multiple_of: 2}, ...&gt; {:integer, multiple_of: 3} ...&gt; ]) iex&gt; 0..9 |&gt; Enum.map(&amp;Xema.valid?(any, &amp;1)) |&gt; Enum.with_index() [true: 0, false: 1, true: 2, true: 3, true: 4, false: 5, true: 6, false: 7, true: 8, true: 9] With one_of exactly on schema have to match. iex&gt; one = Xema.new(one_of: [ ...&gt; {:integer, multiple_of: 2}, ...&gt; {:integer, multiple_of: 3} ...&gt; ]) iex&gt; 0..9 |&gt; Enum.map(&amp;Xema.valid?(one, &amp;1)) |&gt; Enum.with_index() [false: 0, false: 1, true: 2, true: 3, true: 4, false: 5, false: 6, false: 7, true: 8, true: 9]"},{"ref":"usage.html#structuring-a-schema","title":"Usage -  Structuring a schema","type":"extras","doc":"This section will present some options to structuring and reuse schemas. definitions and ref To reuse a schema put it under the keyword definitions. Later on, the schema can be referenced with the keyword ref. iex&gt; schema = Xema.new { ...&gt; :map, ...&gt; definitions: %{ ...&gt; positive: {:integer, minimum: 1}, ...&gt; negative: {:integer, maximum: -1} ...&gt; }, ...&gt; properties: %{ ...&gt; a: {:ref, &quot;#/definitions/positive&quot;}, ...&gt; b: {:ref, &quot;#/definitions/positive&quot;}, ...&gt; c: {:ref, &quot;#/definitions/negative&quot;}, ...&gt; # You can also reference any other schema. ...&gt; d: {:ref, &quot;#/properties/c&quot;} ...&gt; } ...&gt; } ...&gt; Xema.validate schema, %{a: 1, c: -1} :ok iex&gt; {:error, error} = Xema.validate schema, %{b: 1, c: 1} {:error, %Xema.ValidationError{ reason: %{properties: %{c: %{maximum: -1, value: 1}}} }} iex&gt; Exception.message(error) &quot;Value 1 exceeds maximum value of -1, at [:c].&quot; iex&gt; Xema.validate schema, %{d: -1} :ok iex&gt; {:error, error} = Xema.validate schema, %{d: 1} {:error, %Xema.ValidationError{ reason: %{properties: %{d: %{maximum: -1, value: 1}}} }} iex&gt; Exception.message(error) &quot;Value 1 exceeds maximum value of -1, at [:d].&quot; Without definitions and ref It is also possible to use a schema in another schema, as in the following code. iex&gt; positive = Xema.new {:integer, minimum: 1} ...&gt; negative = Xema.new {:integer, maximum: -1} ...&gt; schema = Xema.new { ...&gt; :map, ...&gt; properties: %{ ...&gt; a: positive, ...&gt; b: positive, ...&gt; c: negative ...&gt; } ...&gt; } ...&gt; Xema.validate schema, %{a: 1, b: 2, c: -3} :ok In multiple files To structure schemas in multiple files you have to configure a loader to laod the files. The section &quot;Configure a loader&quot; described the configuration and implementation of an loader. Let&#39;s assume you have the following file available at https://localhost:1234/positive.exon. { :string, definitions: %{ or_nil: {:any_of, [:nil, {:ref, &quot;#&quot;}]} } } I admit this example is a little absurd. But the URL could be also any other URL and with another loader, the file could be on your hard disk. You can use the schema above as follow. iex&gt; schema = Xema.new { ...&gt; :map, ...&gt; id: &quot;http://localhost:1234&quot;, ...&gt; properties: %{ ...&gt; name: {:ref, &quot;xema_name.exon#/definitions/or_nil&quot;}, ...&gt; str: {:ref, &quot;xema_name.exon&quot;} ...&gt; } ...&gt; } ...&gt; Xema.validate schema, %{str: &quot;foo&quot;} :ok iex&gt; {:error, error} = Xema.validate schema, %{str: nil} {:error, %Xema.ValidationError{ reason: %{properties: %{str: %{type: :string, value: nil}}} }} iex&gt; Exception.message(error) &quot;Expected :string, got nil, at [:str].&quot; iex&gt; Xema.validate schema, %{name: nil} :ok iex&gt; Xema.validate schema, %{name: &quot;Funny van Dannen&quot;} :ok iex&gt; {:error, error} = Xema.validate schema, %{name: 66} {:error, %Xema.ValidationError{ reason: %{ properties: %{ name: %{any_of: [ %{type: nil, value: 66}, %{type: :string, value: 66} ], value: 66} } } } } iex&gt; Exception.message(error) &quot;&quot;&quot; No match of any schema, at [:name]. Expected nil, got 66, at [:name]. Expected :string, got 66, at [:name].\\ &quot;&quot;&quot;"},{"ref":"cast.html","title":"Cast","type":"extras","doc":"Cast Xema provides the functions Xema.cast/2 and Xema.cast!/2 to convert data according to the schema. The converted data will also be validated against the schema. The function Xema.cast/2 returns an error tuple with a CastError when the data cannot be converted. iex&gt; schema = Xema.new({:integer, minimum: 1}) iex&gt; Xema.cast(schema, &quot;5&quot;) {:ok, 5} iex&gt; Xema.cast(schema, &quot;-5&quot;) {:error, %Xema.ValidationError{ reason: %{minimum: 1, value: -5} }} iex&gt; {:error, error} = Xema.cast(schema, []) {:error, %Xema.CastError{ key: nil, path: [], to: :integer, value: [] }} iex&gt; Exception.message(error) &quot;cannot cast [] to :integer&quot; With use Xema the functions are also available. iex&gt; defmodule Int do ...&gt; use Xema ...&gt; ...&gt; xema do: integer(minimum: 1) ...&gt; end iex&gt; iex&gt; Int.cast(&quot;6&quot;) {:ok, 6} iex&gt; Int.cast(&quot;-1&quot;) {:error, %Xema.ValidationError{ reason: %{minimum: 1, value: -1} }}"},{"ref":"cast.html#maps","title":"Cast - Maps","type":"extras","doc":"To convert map keys the schema has to specify the keys keyword. iex&gt; defmodule MapSchema do ...&gt; use Xema ...&gt; ...&gt; xema do ...&gt; map( ...&gt; keys: :atoms, ...&gt; properties: %{ ...&gt; pos: integer(minimum: 0), ...&gt; neg: integer(maximum: 0) ...&gt; } ...&gt; ) ...&gt; end ...&gt; end iex&gt; iex&gt; MapSchema.cast(%{&quot;pos&quot; =&gt; &quot;5&quot;, &quot;neg&quot; =&gt; &quot;-5&quot;}) {:ok, %{pos: 5, neg: -5}}"},{"ref":"cast.html#caster","title":"Cast - Caster","type":"extras","doc":"To convert &quot;special&quot; structs a caster can be specified. A caster function has to return the converted value in a tuple {:ok, value} or an :error. iex&gt; defmodule UrisSchema do ...&gt; use Xema ...&gt; ...&gt; def caster(string) when is_binary(string), do: {:ok, URI.parse(string)} ...&gt; ...&gt; def caster(%URI{} = uri), do: {:ok, uri} ...&gt; ...&gt; def caster(_), do: :error ...&gt; ...&gt; xema do ...&gt; map( ...&gt; keys: :atoms, ...&gt; properties: %{ ...&gt; uris: list( ...&gt; items: strux(URI, caster: &amp;UrisSchema.caster/1) ...&gt; ) ...&gt; } ...&gt; ) ...&gt; end ...&gt; end iex&gt; UrisSchema.cast(%{uris: [&quot;https://hexdocs.pm/elixir/Kernel.html&quot;]}) {:ok, %{uris: [URI.parse(&quot;https://hexdocs.pm/elixir/Kernel.html&quot;)]}} The caster can also be a tuple of module and function name or an mfa tuple. Using Caster behaviour is also supported. iex&gt; defmodule UriCaster do ...&gt; @behaviour Xema.Caster ...&gt; ...&gt; @impl true ...&gt; def cast(%URI{} = uri), do: {:ok, uri} ...&gt; ...&gt; def cast(string) when is_binary(string), do: {:ok, URI.parse(string)} ...&gt; ...&gt; def cast(_), do: :error ...&gt; end iex&gt; iex&gt; defmodule UriSchema do ...&gt; use Xema ...&gt; ...&gt; xema do ...&gt; map( ...&gt; keys: :strings, ...&gt; properties: %{ ...&gt; &quot;uri&quot; =&gt; strux(URI, caster: UriCaster) ...&gt; } ...&gt; ) ...&gt; end ...&gt; end iex&gt; iex&gt; UriSchema.cast(%{uri: &quot;https://elixir-lang.org/docs.html&quot;}) {:ok, %{&quot;uri&quot; =&gt; URI.parse(&quot;https://elixir-lang.org/docs.html&quot;)}}"},{"ref":"loader.html","title":"Configure a loader","type":"extras","doc":"Configure a loader A loader returns the data for a remote schema. The remote schemas are defined in a schema like this. ... properties: %{ int: {:ref, &quot;http://localhost:1234/int.exon&quot;} } ... A loader will be configured like this. config :xema, loader: My.Loader A loader is a module which use the behaviour Xema.Loader. defmodule My.Loader do @moduledoc false @behaviour Xema.Loader @spec fetch(binary) :: {:ok, map} | {:error, any} def fetch(uri) do with {:ok, response} &lt;- get(uri), do: eval(response, uri) end defp get(uri) do case HTTPoison.get(uri) do {:ok, %HTTPoison.Response{status_code: 200, body: body}} -&gt; {:ok, body} {:ok, %HTTPoison.Response{status_code: 404}} -&gt; {:error, &quot;Remote schema &#39;\#{uri}&#39; not found.&quot;} {:ok, %HTTPoison.Response{status_code: code}} -&gt; {:error, &quot;code: \#{code}&quot;} {:error, reason} -&gt; {:error, reason} end end defp eval(str, uri) do {data, _} = Code.eval_string(str) {:ok, data} rescue error -&gt; {:error, %{error | file: URI.to_string(uri)}} end end The function fetch/1 will be called by Xema and expects an %URI{}. The return value must be a tuple of :ok and the required data for a schema or an error tuple. Note! This loader use Code.eval_string/1 and eval is always evil. Warning: string can be any Elixir code and will be executed with the same privileges as the Erlang VM: this means that such code could compromise the machine (for example by executing system commands). Don’t use eval_string/3 with untrusted input (such as strings coming from the network). -- Elixir API"},{"ref":"loader.html#file-loader","title":"Configure a loader - File loader","type":"extras","doc":"A loader to read schema from the local file system. In the schema: ... properties: %{ int: {:ref, &quot;int.exon&quot;} } ... The loader: defmodule My.Loader do @moduledoc false @behaviour Xema.Loader @spec fetch(binary) :: {:ok, map} | {:error, any} def fetch(uri), do: &quot;path/to/schemas&quot; |&gt; Path.join(uri.path) |&gt; File.read!() |&gt; eval(uri) defp eval(str, uri) do {data, _} = Code.eval_string(str) {:ok, data} rescue error -&gt; {:error, %{error | file: URI.to_string(uri)}} end end"},{"ref":"examples.html","title":"Examples","type":"extras","doc":"Examples A bunch of Xema examples."},{"ref":"examples.html#basic","title":"Examples - Basic","type":"extras","doc":"A minimal example. iex&gt; defmodule Example.Basic do ...&gt; use Xema, multi: true ...&gt; ...&gt; xema :person do ...&gt; map( ...&gt; properties: %{ ...&gt; first_name: :string, ...&gt; last_name: :string, ...&gt; age: {:integer, minimum: 0} ...&gt; } ...&gt; ) ...&gt; end ...&gt; ...&gt; @default true ...&gt; xema :foo, do: :string ...&gt; end iex&gt; iex&gt; Example.Basic.valid?( ...&gt; :person, ...&gt; %{first_name: &quot;James&quot;, last_name: &quot;Brown&quot;, age: 42} ...&gt; ) true ...&gt; Example.Basic.valid?( ...&gt; :person, ...&gt; %{first_name: :james, last_name: &quot;Brown&quot;, age: 42} ...&gt; ) false iex&gt; {:error, error} = Example.Basic.validate( ...&gt; :person, ...&gt; %{first_name: :james, last_name: &quot;Brown&quot;, age: 42} ...&gt; ) {:error, %Xema.ValidationError{ reason: %{properties: %{first_name: %{type: :string, value: :james}}} }} iex&gt; Exception.message(error) &quot;Expected :string, got :james, at [:first_name].&quot; iex&gt; Example.Basic.valid?(&quot;foo&quot;) true iex&gt; Example.Basic.valid?(:foo) false"},{"ref":"examples.html#options","title":"Examples - Options","type":"extras","doc":"An example to check opts. iex&gt; defmodule Example.Options do ...&gt; use Xema ...&gt; ...&gt; xema do ...&gt; keyword( ...&gt; properties: %{ ...&gt; foo: atom(enum: [:bar, :baz]), ...&gt; limit: integer(minimum: 0), ...&gt; msg: :string ...&gt; }, ...&gt; required: [:foo, :limit], ...&gt; additional_properties: false ...&gt; ) ...&gt; end ...&gt; end iex&gt; iex&gt; Example.Options.validate(foo: :bar, limit: 11, msg: &quot;foo&quot;) :ok iex&gt; {:error, error} = Example.Options.validate(foo: :foo, limit: 11) {:error, %Xema.ValidationError{ reason: %{properties: %{foo: %{enum: [:bar, :baz], value: :foo}}} }} iex&gt; Exception.message(error) &quot;Value :foo is not defined in enum, at [:foo].&quot; iex&gt; {:error, error} = Example.Options.validate(foo: :bar) {:error, %Xema.ValidationError{ reason: %{required: [:limit]} }} iex&gt; Exception.message(error) &quot;Required properties are missing: [:limit].&quot; iex&gt; {:error, error} = Example.Options.validate(foo: :bar, limit: 11, message: &quot;foo&quot;) {:error, %Xema.ValidationError{ reason: %{properties: %{message: %{additional_properties: false}}} }} iex&gt; Exception.message(error) &quot;Expected only defined properties, got key [:message].&quot;"},{"ref":"examples.html#custom-validator","title":"Examples - Custom validator","type":"extras","doc":"This example shows the use of an custom validator that is given as an tuple of module and function name. iex&gt; defmodule Example.Palindrome do ...&gt; def check(str) do ...&gt; case str == String.reverse(str) do ...&gt; true -&gt; :ok ...&gt; false -&gt; {:error, :no_palindrome} ...&gt; end ...&gt; end ...&gt; end iex&gt; iex&gt; defmodule Example.PaliSchema do ...&gt; use Xema ...&gt; ...&gt; xema :palindrome do ...&gt; string(validator: {Example.Palindrome, :check}) ...&gt; end ...&gt; end ...&gt; ...&gt; Example.PaliSchema.valid?(:palindrome, &quot;racecar&quot;) true iex&gt; Example.PaliSchema.valid?(:palindrome, &quot;bike&quot;) false iex&gt; {:error, error} = Example.PaliSchema.validate(:palindrome, &quot;bike&quot;) {:error, %Xema.ValidationError{ reason: %{validator: :no_palindrome, value: &quot;bike&quot;} }} iex&gt; Exception.message(error) ~s|Validator fails with :no_palindrome for value &quot;bike&quot;.| A validator can also be specified as behaviour. iex&gt; defmodule Example.PalindromeB do ...&gt; @behaviour Xema.Validator ...&gt; ...&gt; @impl true ...&gt; def validate(str) do ...&gt; case str == String.reverse(str) do ...&gt; true -&gt; :ok ...&gt; false -&gt; {:error, :no_palindrome} ...&gt; end ...&gt; end ...&gt; end iex&gt; iex&gt; defmodule Example.PaliSchemaB do ...&gt; use Xema ...&gt; ...&gt; xema :palindrome do ...&gt; string(validator: Example.PalindromeB) ...&gt; end ...&gt; end ...&gt; ...&gt; Example.PaliSchemaB.valid?(:palindrome, &quot;racecar&quot;) true iex&gt; Example.PaliSchemaB.valid?(:palindrome, &quot;bike&quot;) false iex&gt; {:error, error} = Example.PaliSchemaB.validate(:palindrome, &quot;bike&quot;) {:error, %Xema.ValidationError{ reason: %{validator: :no_palindrome, value: &quot;bike&quot;} }} iex&gt; Exception.message(error) ~s|Validator fails with :no_palindrome for value &quot;bike&quot;.| The custom validator can also be a part of the schema module. iex&gt; defmodule Example.Range do ...&gt; use Xema ...&gt; ...&gt; xema :range do ...&gt; map( ...&gt; properties: %{ ...&gt; from: integer(minimum: 0), ...&gt; to: integer(maximum: 100) ...&gt; }, ...&gt; validator: &amp;Example.Range.check/1 ...&gt; ) ...&gt; end iex&gt; iex&gt; def check(%{from: from, to: to}) do ...&gt; case from &lt; to do ...&gt; true -&gt; :ok ...&gt; false -&gt; {:error, :from_greater_to} ...&gt; end ...&gt; end ...&gt; end ...&gt; ...&gt; Example.Range.validate(:range, %{from: 6, to: 8}) :ok iex&gt; {:error, error} = Example.Range.validate(:range, %{from: 66, to: 8}) {:error, %Xema.ValidationError{ reason: %{validator: :from_greater_to, value: %{from: 66, to: 8}} }} iex&gt; Exception.message(error) &quot;Validator fails with :from_greater_to for value %{from: 66, to: 8}.&quot; iex&gt; {:error, error} = Example.Range.validate(:range, %{from: 166, to: 118}) {:error, %Xema.ValidationError{ reason: %{properties: %{to: %{maximum: 100, value: 118}}} }} iex&gt; Exception.message(error) &quot;Value 118 exceeds maximum value of 100, at [:to].&quot;"},{"ref":"examples.html#cast-json","title":"Examples - Cast JSON","type":"extras","doc":"The following example cast data structure that is decoded by Jason.decode!/1. For the encoding of URI a Jason.Encoder implementation is needed. defimpl Jason.Encoder, for: URI do def encode(uri, _opts) do ~s|&quot;\#{URI.to_string(uri)}&quot;| end end This example shows how a complex data structure decoded by a JSON parser can be converted in a form described by a schema. iex&gt; defmodule CasterUri do ...&gt; @behaviour Xema.Caster ...&gt; ...&gt; @impl true ...&gt; def cast(%URI{} = uri), do: {:ok, uri} ...&gt; ...&gt; def cast(string) when is_binary(string), do: {:ok, URI.parse(string)} ...&gt; ...&gt; def cast(_), do: :error ...&gt; end iex&gt; iex&gt; defmodule UserSchema do ...&gt; use Xema ...&gt; ...&gt; xema :user do ...&gt; map( ...&gt; keys: :atoms, ...&gt; properties: %{ ...&gt; name: :string, ...&gt; birthday: strux(Date), ...&gt; favorites: ...&gt; map( ...&gt; keys: :atoms, ...&gt; properties: %{ ...&gt; fruits: list(items: atom(enum: [:apple, :orange, :banana])), ...&gt; uris: list(items: strux(URI, caster: CasterUri)) ...&gt; } ...&gt; ) ...&gt; }, ...&gt; additional_properties: false ...&gt; ) ...&gt; end ...&gt; end iex&gt; iex&gt; {:ok, json} = ...&gt; %{ ...&gt; &quot;name&quot; =&gt; &quot;Nick&quot;, ...&gt; &quot;birthday&quot; =&gt; ~D|2000-04-17|, ...&gt; &quot;favorites&quot; =&gt; %{ ...&gt; &quot;fruits&quot; =&gt; ~w(apple banana), ...&gt; &quot;uris&quot; =&gt; [&quot;https://elixir-lang.org/&quot;] ...&gt; } ...&gt; } ...&gt; |&gt; UserSchema.cast!() ...&gt; |&gt; Jason.encode() {:ok, &quot;{\\&quot;birthday\\&quot;:\\&quot;2000-04-17\\&quot;,\\&quot;favorites\\&quot;:{\\&quot;fruits\\&quot;:[\\&quot;apple\\&quot;,\\&quot;banana\\&quot;],\\&quot;uris\\&quot;:[\\&quot;https://elixir-lang.org/\\&quot;]},\\&quot;name\\&quot;:\\&quot;Nick\\&quot;}&quot;} iex&gt; iex&gt; json |&gt; Jason.decode!() |&gt; UserSchema.cast!() %{ birthday: ~D[2000-04-17], favorites: %{ fruits: [:apple, :banana], uris: [ %URI{ authority: &quot;elixir-lang.org&quot;, fragment: nil, host: &quot;elixir-lang.org&quot;, path: &quot;/&quot;, port: 443, query: nil, scheme: &quot;https&quot;, userinfo: nil } ] }, name: &quot;Nick&quot; }"},{"ref":"unsupported.html","title":"Unsupported Features","type":"extras","doc":"Unsupported Features The keywords readOnly and writeOnly are not supported. The keyword contentMediaType and contentEncoding are just for annotation and will not be used for validation. The following format checks are not supported: idn-email the I18N equivalent of email idn-hostname I18N equivalent of hostname iri the I18N equivalent of uri iri-reference the I18N equivalent of uri-reference"}]